
/* program:  sds2phdu.c
 * function: Convert sds file to fits' Primary array  
 *
 *
 *
 */

#include <stdio.h> 
#include <string.h>
#include <stdlib.h>

/* Package include files */
#include "fits2sds.h"
#include "extern.h"
#include "cfitsio.h"

/* HDF include files */
#include "hdf.h"
#include "herr.h"

#define MAX_NC_NAME 256          /* max length of a name */  
#define MAX_VAR_DIMS 32          /* max per variable dimensions */

int sds2phdu(sd_id, unitNum)
    int unitNum,sd_id;
{
    /* some of fits variable */
    int simple,extend,anyflg, iounit=unitNum;
    int numberKeys, numberSpace;
    int bitpix,naxis,naxes[15],hdfNaxes[15], pcount,gcount;
    int nelmt;
    int group,fpixel;
    char keyword[8];
    char value[24], comment[70];


    /* some of sds variable */
    int32	sds_id, dim_id;
    int32	numberType, sdsDataType;
    int32	start[15], edges[15];
    int32       sds_index, dim_index;
    int32       rank, numAttr,numSdsAttr, numDataset, count;
    int32       dimSizes[MAX_VAR_DIMS];
    char        name[MAX_NC_NAME], sdsName[MAX_NC_NAME];
    int8        *buffer;
    char        label[12],unit[12],format[12];

    /* define array to hold attr info */
    sdsAttr  sdsAttrRec[999];
    fitsHead fitsHeadRec[999];


    /* pointer of various types */
    union {
	char	*cp;	/* Byte or char */
	short int	*sp;	/* short */
	int	*ip;	/* integer */
	float32	*fp;	/* float */
	float64	*dp;	/* double */
    } databuf;
    
    float32  *databuf_def;

    int datasize;	/* in bytes */

    int i, j;
    int nbytes;
    int status;
    
    uint8 *ptr,*sptr,*dataPtr;
    int keywordIndex;
    int keywordLen;

    float32 bscaleVal;
    float64 sdsBscaleVal;
    int32 range[2];
    int16 fitsBitspix[1];
    int32 dim_num;

    float64 dimScale[5000];


    char nextKey[8];
    char card[80];
    int commentCount = 0, blankCount = 0; /*  */
    int historyCount = 0;
    int fillVal;
    float32 dataMax, dataMin;
    float32 crpixVal, crotaVal, crvalVal, cdeltVal;

    char commentAttr[8],commentVal[1000];
    char historyAttr[8],historyVal[1000];
    char naxisComment[16];
    
    int  datamaxFlag=0, dataminFlag=0;
    int  bscaleFlag,bzeroFlag;
    int32 bscaleFactor=1;
    
    int32 shiftDataFlag=0;   
    
    char errtxt[FITS_CLEN_ERRMSG];
    
    /* more than 2 sds */
    int firstSdsFlag = 1;     /* suppose only one sds */
    int numSds = 0;           /* real sds number in hdf file */


    /* get sds file information  */
    status = SDfileinfo(sd_id, &numDataset, &numAttr);

    /* get the identifier of the first data set  */
    sds_index = 0;
    /* get real sds number of HDF file */
    while (sds_index < numDataset) {
      sds_id = SDselect(sd_id, sds_index);
      if (FALSE == SDiscoordvar(sds_id))
	++numSds;
      ++sds_index;
    }

    sds_index=0;
    sds_id = SDselect(sd_id, sds_index);
 
    /* get the information about this dataset  */
    while (sds_index < numDataset ) {

      bscaleFlag = 0;
      bzeroFlag  = 0;
      numSdsAttr = -1;

      if (FALSE == SDiscoordvar(sds_id)) { /* dataset is not a coordinate variable  */
	status = SDgetinfo(sds_id, name , & naxis, hdfNaxes,  &numberType, &numAttr);
	
	/* keep sds's data type */
	sdsDataType = numberType;

	/* check  sds number type to see if need to shift data  */
	if ((sdsDataType==DFNT_INT8)||(sdsDataType==DFNT_UINT16)||(sdsDataType==DFNT_CHAR16)||(sdsDataType==DFNT_UINT32)||(sdsDataType==DFNT_CHAR8)) 
	    shiftDataFlag = 1;

	/* keep sds's name */
	strcpy(sdsName, name);
	
	/* reverse naxes */
	for (i=0; i<naxis; i++) 
	  naxes[i] = hdfNaxes[naxis-i-1];
	
	/* get all of attributes  */
	j=-1;
	for (i=0; i<numAttr; i++) {
	  /* get information about the dataset attribute  */
	  status = SDattrinfo(sds_id, i, name, &numberType, &count);
	  
  	  if (strncmp(name,"_FillValue",10) == 0) 
	    strcpy(name,"BLANK");
	  
	  if (strncmp(name,"valid_max",9) == 0) 
	    strcpy(name,"DATAMAX");
	  
	  if (strncmp(name,"valid_min",9) == 0) 
	    strcpy(name,"DATAMIN");

	  if (strncmp(name,"units",5) == 0)
	    strcpy(name,"BUNIT");
	  /* add keywords DATALABL DATAFMT DATACOOR  when get data string info */
	  if (strcmp(name,"long_name") == 0)
	    strcpy(name,"DATALABL");
	  if (strcmp(name,"format")== 0)
	    strcpy(name,"DATAFMT");
	  if (strcmp(name,"cordsys")==0)
	    strcpy(name,"DATACOOR");

	  /* dimension scale */
	  if (strcmp(name,"add_offset")==0)
	    strcpy(name, "BZERO");
	  if (strcmp(name,"scale_factor")==0)
	    strcpy(name, "BSCALE");

	  if (strcmp(name,"add_offset_err")==0)
	    strcpy(name, "BZEROERR");
	  if (strcmp(name,"scale_factor_err")==0)
	    strcpy(name, "BSCALERR");

	  /* Allocate a buffer to hold the attribute data  */
	  buffer = (int8 *)malloc(count*DFKNTsize(numberType));
		
	  /* read the attribute data  */
	  status = SDreadattr(sds_id, i, buffer);
	  
	  
	  if (strcmp(name,"valid_range")&&(numberType!=DFNT_CHAR8)&&(count>=2)) {
	    j=convSdsGenericAttr(sdsAttrRec,j,name,numberType,buffer,count);
	    continue;
	  }

		
	  if (strcmp(name,"calibrated_nt")) {  /*  get rid of this */ 

	    if (strcmp(name,"valid_range")) {
	      ++j;
	      if (strcmp(name,"BSCALERR") && strcmp(name,"BZEROERR") && strcmp(name,"BLANK")) 
		/* write info to sdsAttrRec that hold fits keyword info  */
		writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
	      else {

		if ((strcmp(name,"BLANK")==0)&&(numberType!=DFNT_FLOAT32)&&(numberType!=DFNT_FLOAT64))
		  /* write info to sdsAttrRec that hold fits keyword info  */
		  writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count); 
		{ /* specil proceeding */
		  float64 *tmp = (float64 *)buffer;
		  if ((strcmp(name,"BSCALERR")==0)&&(numberType != DFNT_UINT8))
		  if (*tmp != 0.0) {
		    bscaleFlag = 1;
		    /* write info to sdsAttrRec that hold fits keyword info  */
		    writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
		  }
		  else { /* bscalerr == 0 */
		    if (bscaleFlag)
		      /* write info to sdsAttrRec that hold fits keyword comment info  */
		      writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
		  }

		if ((strcmp(name,"BZEROERR")==0)&&(numberType != DFNT_UINT8))
		  if (*tmp != 0.0) {
		    bzeroFlag = 1;
		    /* write info to sdsAttrRec that hold fits keyword info  */
		    writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
		  }
		  else { /* bzeroerr == 0 */
		    if (bzeroFlag)
		      /* write info to sdsAttrRec that hold fits keyword info  */
		      writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
		  }
		} /* specil proceeding */
	      }
	      
	    } /* if (strcmp(name,"valid_range")) { */
	    else {
	      ++j;
	      strcpy(name,"DATAMIN");
	      writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
	      strcpy(name,"DATAMAX");
	      datasize = getBitpix(numberType);
	      datasize = (datasize<0) ? -datasize : datasize;
	      datasize = datasize/8;
	      buffer    = buffer + datasize;
	      ++j;
	      writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
	    }
	  } /* if strcmp(name,"calibrated_nt").... */
	  else {
	    int32 *calType=(int32 *)buffer;
	    strcpy(name,"DATATYPE");
	    if (*calType != DFNT_FLOAT32) { 
	      ++j;
	      /*write info to sdsAttrRec that hold fits keyword comment info */
	      writeSdsAttrRec(sdsAttrRec, j, name, numberType, buffer, count);
	    }
	  }

	} /* for i<numAttr  */
	
	numSdsAttr = numSdsAttr + j + 1;   /* j: 0~j */


	
	    /* read dimension scale */
	    j=1;
	    for (dim_index=naxis-1; dim_index>=0 ; dim_index--) {  /* reverse */

	      /* get a dimension identifier for a specified dimension of an SDS. */
	      dim_id = SDgetdimid(sds_id,dim_index);
	      
	      if (FAIL != dim_id) { /* SUCCEED */
		/* if dimension not to set return numberType that should not be set ( the value is zero)  */
		status = SDdiminfo(dim_id, name , &count, &numberType, &numAttr);
		/* Do we need to convert dimension name */
		/* How about dimension Number Type? */

		if (numberType>0) { 

		/* read dimension scale value */
		status = SDgetdimscale(dim_id,dimScale);
		if (status == SUCCEED) {
		  
		  keywordIndex = naxis - dim_index;
		  /* read succeedly */
		  strcpy(name,"CRVAL");
		  strcat(name,num2str(keywordIndex));
		  crvalVal = dimScale[1];
		  writeSdsAttrRec(sdsAttrRec, numSdsAttr+j, name,DFNT_FLOAT32, &crvalVal, 1);

		  strcpy(name,"CRPIX");
		  strcat(name,num2str(keywordIndex));
		  crpixVal=1.0;
		  writeSdsAttrRec(sdsAttrRec, numSdsAttr+j+1, name,DFNT_FLOAT32, &crpixVal , 1);
		  strcpy(name,"CDELT");
		  strcat(name,num2str(keywordIndex));
		  cdeltVal = dimScale[1]-dimScale[0];
		  writeSdsAttrRec(sdsAttrRec, numSdsAttr+j+2, name,DFNT_FLOAT32, &cdeltVal , 1);
		  /* strcpy(name,"CROTA");
		  strcat(name,num2str(keywordIndex));
		  crotaVal = 0.0;
		  writeSdsAttrRec(sdsAttrRec, numSdsAttr+j+3, name,DFNT_FLOAT32, &crotaVal , 1);
		  */
		}
		j = j+3;

		/* CTYPEn  */
		strcpy(label,"");
		strcpy(unit,"");
		strcpy(format,"");

		status = SDgetdimstrs(dim_id, label,unit,format,12);
		if (status == SUCCEED) {
		  strcpy(name,"CTYPE");
		  strcat(name,num2str(keywordIndex));
		  writeSdsAttrRec(sdsAttrRec, numSdsAttr+j, name,DFNT_CHAR8, unit , strlen(unit));
		  ++j;
		  /* add keywords CLABLn CFMTn */
		  if (strlen(trim(label))) {
		    strcpy(name,"CLABL");
		    strcat(name,num2str(keywordIndex));
		    writeSdsAttrRec(sdsAttrRec, numSdsAttr+j, name,DFNT_CHAR8,label, strlen(label));
		    ++j;
		  }
		  if (strlen(trim(format))) {
		    strcpy(name,"CFMT");
		    strcat(name,num2str(keywordIndex));
		    writeSdsAttrRec(sdsAttrRec, numSdsAttr+j, name,DFNT_CHAR8,format, strlen(format));
		    ++j;
		  }
		} /* 	if (status == SUCCEED) { */
		} /* if (numAttr >0) */
	      } /* if (FAIL==dim_id) */
	    } /*  for (dim_index=naxis-1; dim_index>=0 ; dim_index--) { */


	    numSdsAttr = numSdsAttr + j -1;
	    j=0;
	    for (dim_index=naxis-1; dim_index>=0; dim_index--) {  /* reverse */

		/* get a dimension identifier for a specified dimension of an SDS. */
		dim_id = SDgetdimid(sds_id,dim_index);
	    
		if (FAIL != dim_id) { /* SUCCEED */

		  /* get dimension information */
		  status = SDdiminfo(dim_id, name , &count, &numberType, &numAttr);
		    keywordIndex = naxis - dim_index;
		    for (i=0; i<numAttr; i++) {
			/* get info about selected dimension  */
			status = SDattrinfo(dim_id, i, name, &numberType, &count);
		     
			if (strncmp(name,"units",5) && strcmp(name,"long_name")&&strcmp(name,"format")) {  /* skip this attrs */
			
			  /* Allocate a buffer to hold the attribute data  */
			  buffer = (int8 *)malloc(count*DFKNTsize(numberType));
			
			  /* read the attribute data  */
			  status = SDreadattr(dim_id, i, buffer);
			  
			  /* write info to sdsAttrRec that hold fits keyword info  */
			  ++j;
			  writeSdsAttrRec(sdsAttrRec, numSdsAttr+j, name, numberType, buffer, count);
			} /* if (strncmp(name,"units",5) || */
		    } /* for (i=0; i<numAttr; i++) */
		} /* if (FAIL != dim_id) { */
	    }  /* for (dim_index=0; dim_index<naxis; dim_index++) {  */

	    numSdsAttr +=j;

	    /* if sdsName is '<EXTNAME>_<EXTVER>' keep fits keyword EXTNAME value */
	    if ((at("FITS_PHDU",sdsName)<0)&&(at("FITS_IMAGE",sdsName)<0)) {
	      numSdsAttr++;

	      /* see if EXTVER is involved */
	      if (at("_", sdsName)>0) {
		strcpy(name,substr(sdsName,1,at("_", sdsName)));
		writeSdsAttrRec(sdsAttrRec, numSdsAttr, "EXTNAME",DFNT_CHAR8,name,strlen(name));
		numSdsAttr++;
		strcpy(name,substr(sdsName,at("_", sdsName)+2, 2));
		strcpy(sdsAttrRec[numSdsAttr].keyword, "EXTVER");
		sdsAttrRec[numSdsAttr].dataType = DFNT_INT32;
		sdsAttrRec[numSdsAttr].iKeywordVal = atoi(name);
	      }
	      else
 
		writeSdsAttrRec(sdsAttrRec, numSdsAttr, "EXTNAME",DFNT_CHAR8,sdsName,strlen(sdsName));
	    }

	    /* initinalizing array  fitsHeadRec */
	    initArray(fitsHeadRec,numSdsAttr);


	/* convert sdsAttrRec to standard card so as to write fits head */
	count = 0;
	for (i=0; i<=numSdsAttr; i++) {

	  if ((keywordLen=at("COMMENT",sdsAttrRec[i].keyword)) > 0) { /* this is a comment form */
	    strcpy(keyword, substr(sdsAttrRec[i].keyword, 1, keywordLen-1));
	    keywordIndex = lookupKeyword(fitsHeadRec, keyword, count);
	    if (keywordIndex >= 0)  /* found keyword  */		
	      strcpy(fitsHeadRec[keywordIndex].keywordComm, sdsAttrRec[i].cKeywordVal);
	    else { /* datamax_comment, ... */	        
	      strcpy(fitsHeadRec[count].keyword,keyword);
	      fitsHeadRec[count].dataType = sdsAttrRec[i].dataType;
	      strcpy(fitsHeadRec[count].keywordComm, sdsAttrRec[i].cKeywordVal);
	      ++count;
	    } 
	  }
	  else {  /*  attributes to be written (comment1 history1 blank1 orgin */
  
	    strcpy(keyword , sdsAttrRec[i].keyword);
	    if (sdsAttrRec[i].dataType != DFNT_UINT8) { /* such like bitpix ... */
	      keywordIndex = lookupKeyword(fitsHeadRec, keyword, count);  /* find DATAMIN ... */
	      if (keywordIndex <= 0) { /* not Found */
		
		strcpy(fitsHeadRec[count].keyword, keyword);
		fitsHeadRec[count].dataType    = sdsAttrRec[i].dataType;
		switch(sdsAttrRec[i].dataType) {
		case DFNT_INT16: /* short */
		    fitsHeadRec[count].sKeywordVal = sdsAttrRec[i].sKeywordVal;
		    break;
		case DFNT_INT32: /* int */
		    fitsHeadRec[count].iKeywordVal = sdsAttrRec[i].iKeywordVal;
		    break;
		case DFNT_FLOAT32: /* float32  */
		    fitsHeadRec[count].fKeywordVal = sdsAttrRec[i].fKeywordVal;
		    break;

		case DFNT_FLOAT64: /* float64  */
		    fitsHeadRec[count].fKeywordVal = sdsAttrRec[i].fKeywordVal;
		    break;
		}
		++count;
	      }
	      else {
		fitsHeadRec[keywordIndex].dataType    = sdsAttrRec[i].dataType;
		switch(sdsAttrRec[i].dataType) {
		case DFNT_INT16: /* short */
		    fitsHeadRec[keywordIndex].sKeywordVal = sdsAttrRec[i].sKeywordVal;
		    break;
		case DFNT_INT32: /* int */
		    fitsHeadRec[keywordIndex].iKeywordVal = sdsAttrRec[i].iKeywordVal;
		    break;
		case DFNT_FLOAT32: /* float32  */
		    fitsHeadRec[keywordIndex].fKeywordVal = sdsAttrRec[i].fKeywordVal;
		    break;

		case DFNT_FLOAT64: /* float64  */
		    fitsHeadRec[keywordIndex].fKeywordVal = sdsAttrRec[i].fKeywordVal;
		    break;
		}
	      }  /* if keywordIndex  < 0  */
	      if (strncmp(keyword,"BITPIX",5)==0) { /* BITPIX */
		/* append keyword NAXIS */
		  
		strcpy(fitsHeadRec[count].keyword, "NAXIS");
		fitsHeadRec[count].dataType    = DFNT_INT32;
		fitsHeadRec[count].iKeywordVal = naxis;
		++count;
		for (j=1; j<=naxis; j++) {
		  strcpy(keyword,"NAXIS");
		  strcat(keyword,num2str(j));
		  strcpy(fitsHeadRec[count].keyword, keyword);
		  fitsHeadRec[count].dataType    = DFNT_INT32;
		  fitsHeadRec[count].iKeywordVal = naxes[j-1]; 
		  ++count;
		}
	      }
	    } /* if (sdsAttrRec[i].dataType != DFNT_UINT8) {  */
	    
	    else {   /* comment.... simple ....*/
	      if ((at("COMMENT",keyword)>=0)||(at("HISTORY",keyword)>=0)||(at("blank",keyword)>=0)) { /* comment1 .... */
		char *tmpStr = sdsAttrRec[i].cKeywordVal;
		int  tmpStrLen ,commentLen;
		
		tmpStrLen = strlen(tmpStr);
		
		if (strncmp("COMMENT",keyword,7)==0) {
		  strcpy(keyword, "COMMENT");
		  
		  commentLen = at("\n",tmpStr);
		  while  (commentLen >= 0) {
			  strcpy(comment, substr(tmpStr,1,commentLen));
			  tmpStr += commentLen;
			  ++tmpStr;
			  commentLen = at("\n",tmpStr);
			  strcpy(fitsHeadRec[count].keyword, keyword);
			  fitsHeadRec[count].dataType    = DFNT_UINT8;
			  strcpy(fitsHeadRec[count].keywordComm,comment);
			  ++count;
		  } /* 	while (commentLen > 0) { */
		  strcpy(fitsHeadRec[count].keyword, keyword);
		  fitsHeadRec[count].dataType    = DFNT_UINT8;
		  strcpy(fitsHeadRec[count].keywordComm, tmpStr);
		  ++count;
		} /* if (strncmp("COMMENT",keyword,7)==0) {  */


		if (strncmp("HISTORY",keyword,7)==0) {
		  strcpy(keyword, "HISTORY");
		  
		  commentLen = at("\n",tmpStr);
		  while  (commentLen >= 0) {
		    strcpy(comment, substr(tmpStr,1,commentLen));
		    tmpStr += commentLen;
		    ++tmpStr;
		    commentLen = at("\n",tmpStr);
		    strcpy(fitsHeadRec[count].keyword, keyword);
		    fitsHeadRec[count].dataType    = DFNT_UINT8;
		    strcpy(fitsHeadRec[count].keywordComm,comment);
		    ++count;
		  } /* 	while (commentLen > 0) { */
		  strcpy(fitsHeadRec[count].keyword, keyword);
		  fitsHeadRec[count].dataType    = DFNT_UINT8;
		  strcpy(fitsHeadRec[count].keywordComm, tmpStr);
		  ++count;
		} /* if (strncmp("HISTORY",keyword,7)==0) {  */
		
		
		if (strncmp("blank",keyword,5)==0) {
		  strcpy(keyword, "blank");
		  
		  commentLen = at("\n",tmpStr);
		  while  (commentLen >= 0) {
		    strcpy(comment, substr(tmpStr,1,commentLen));
		    tmpStr += commentLen;
		    ++tmpStr;
		    commentLen = at("\n",tmpStr);
		    strcpy(fitsHeadRec[count].keyword, keyword);
		    fitsHeadRec[count].dataType    = DFNT_UINT8;
		    strcpy(fitsHeadRec[count].keywordComm,comment);
		    ++count;
		  } /* 	while (commentLen > 0) { */
		  strcpy(fitsHeadRec[count].keyword, keyword);
		  fitsHeadRec[count].dataType    = DFNT_UINT8;
		  strcpy(fitsHeadRec[count].keywordComm, tmpStr);
		  ++count;
		} /* if (strncmp("blank",keyword,5)==0) {  */
		
		
	      } /* 	if (at("COMMENT",keyword) ||  ....  */
	      else {  /* simple TITLE  EXTNAME....*/
		keywordIndex = lookupKeyword(fitsHeadRec, keyword, count);
		if (keywordIndex >= 0)  /* found keyword  */		
		  strcpy(fitsHeadRec[keywordIndex].cKeywordVal, sdsAttrRec[i].cKeywordVal);
		else {
		  strcpy(fitsHeadRec[count].keyword, keyword);
		  fitsHeadRec[count].dataType    = DFNT_UINT8;
		  strcpy(fitsHeadRec[count].cKeywordVal,ltrim(sdsAttrRec[i].cKeywordVal));
		  ++count;
		}
	      } 
	    }  /* else {  comment.... simple ....*/
	  } /*  else { bitpix comment ... */
	} /* for (i=0; i<numSdsAttr; i++) */
	

	if (count >= 0) {  /* when count is zero no sds attributes is specified */
	  simple = 1;
	  gcount = 1;
	  pcount = 0;
	  extend = 0;
	  if (numSds>=2)
	    extend = 1;
	  bitpix = getBitpix(sdsDataType);
	  
	  /* FCCRHD(iounit,&status);
	  CHECKFITSERROR("FCCRHD",status,errtxt);
	  */
	  if (count) { 
	    if ((lookupKeyword(fitsHeadRec,"SIMPLE", count) >= 0)||(lookupKeyword(fitsHeadRec,"XTENSION", count) >= 0))

	      /* write fitsHeadRec to fits file */
	      writeFitsHead(iounit, fitsHeadRec, count,sdsDataType);
	    else {  
	      if (firstSdsFlag) { /* the first sds */
		firstSdsFlag = 0;
		/* write the required primary array keywords */
		FCPHPR(iounit, simple, bitpix, naxis, naxes, pcount, gcount, extend, &status);
	      }
	      else {	
		/* create a new empty HDU at end of fits file  */
		FCCRHD(iounit,&status);
		CHECKFITSERROR("FCCRHD",status,errtxt);
		extend = 0;
		/* write the required primary array keywords */
		FCPHPR (iounit, simple, bitpix, naxis, naxes, pcount, gcount, extend, &status);
	      }

	      /* write fitsHeadRec to fits file */
	      writeFitsHead(iounit, fitsHeadRec, count,sdsDataType);
	    }
	  }
	  else {  
	    /* write the required primary array keywords */
	    FCPHPR (iounit, simple, bitpix, naxis, naxes, pcount, gcount, extend, &status);
	    CHECKFITSERROR("FCPHPR",status,errtxt);

	    FCRDEF(iounit,&status);
	    CHECKFITSERROR("FCRDEF",status,errtxt);

	    /* if other data type is defined add BZERO */
	    writeFitsHead(iounit, fitsHeadRec, 0,sdsDataType);
	  }

	  /* 
	  FCRDEF(iounit,&status);
	  CHECKFITSERROR("FCRDEF",status,errtxt);
	  */
	  if ((lookupKeyword(fitsHeadRec,"BSCALE", count) >= 0)||(lookupKeyword(fitsHeadRec,"BZERO", count) >= 0)||(shiftDataFlag)) { 
	  FCPSCL(iounit,1.0,0.0,&status);
	  CHECKFITSERROR("FCPSCL",status,errtxt);
	  }

	  /* Calculate image size and allocate buffer space */
	  nelmt = 1;
	  for (i=0; i<naxis; i++){
	    /* printf("naxes[%d] = %d\n", i+1, naxes[i]); */
	    nelmt *= naxes[i];
	    start[i] = 0;
	  }	
  
	  
	  datasize = (bitpix < 0 ? -bitpix : bitpix)/8 * nelmt;
	  databuf.cp = (char *)newSpace(datasize);
	  group = 0;
	  status = 0;
	 
	  if (naxis>0) { /* Prime data matrix is not empty */
  
	  switch(bitpix) {
	  case 8: /* char */

	    if (shiftDataFlag) {  /* dataType is int8 */

	      int8 *dat_buf;
	      char *ptr = (char *)databuf.cp;

	      dat_buf = (int8 *)newSpace(datasize);

	      SDreaddata(sds_id,start,NULL,hdfNaxes,dat_buf);	     
	      for (i=0;i<nelmt; i++)
		*ptr++ = (*dat_buf++) + 128;
	    }
	    else
	       SDreaddata(sds_id,start,NULL,hdfNaxes,databuf.cp);
	    #if defined(FLIP)  
	    fdat2cdat_uint8(bitpix,naxis,hdfNaxes,databuf.cp);
	    #endif
	    FCPPRB(iounit,group,1,nelmt,(uchar8 *)databuf.cp,&status);
	    break;
	  case 16: {  /* int16 */ 
	
 	    if (shiftDataFlag) {  /* dataType is uint16 */

	      uint16 *dat_buf;
	      int16 *ptr = (int16 *)databuf.sp;

	      dat_buf = (uint16 *)newSpace(datasize);

	      SDreaddata(sds_id,start,NULL,hdfNaxes,dat_buf);
	      for (i=0;i<nelmt; i++)
		*ptr++ = (*dat_buf++) - 32768;
	    }
	    else
	       SDreaddata(sds_id,start,NULL,hdfNaxes,databuf.sp);
	    #if defined(FLIP)
	    fdat2cdat_int16(bitpix,naxis,hdfNaxes,databuf.sp);
	    #endif
	    FCPPRI(iounit,group,1,nelmt,databuf.sp,&status);
	  }
	  break;
	 
	  case 32: /* int */

	    if (shiftDataFlag) {  /* dataType is uint32 */

	      uint32 *dat_buf;
	      int32 *ptr = (int32 *)databuf.ip;

	      dat_buf = (uint32 *)newSpace(datasize);

	      SDreaddata(sds_id,start,NULL,hdfNaxes,dat_buf);
	      for (i=0;i<nelmt; i++)
		*ptr++ = (*dat_buf++) - 2147483648.0;
	    }
	    else
	       SDreaddata(sds_id,start,NULL,hdfNaxes,databuf.ip);
	    #if defined(FLIP)
	    fdat2cdat_int32(bitpix,naxis,hdfNaxes,databuf.ip);
	    #endif
	    FCPPRJ(iounit,group,1,nelmt,databuf.ip,&status);
	    break;

	  case -32: /* float32 */

	    SDreaddata(sds_id,start,NULL,hdfNaxes,databuf.fp);
	    #if defined(FLIP)
	    fdat2cdat_float32(bitpix,naxis,hdfNaxes,databuf.fp);
	    #endif
	    FCPPRE(iounit,group,1,nelmt,databuf.fp,&status);
	    break;
	  case -64: /* float64 */
	    SDreaddata(sds_id,start,NULL,hdfNaxes,databuf.dp);
	    #if defined(FLIP)
	    fdat2cdat_float64(bitpix,naxis,hdfNaxes,databuf.dp);
	    #endif
	    FCPPRD(iounit,group,1,nelmt,databuf.dp,&status);
	    break;
	  default:
	    return(err_msg(WriteFITSFile,BadInputFile));

	  }

	  CHECKFITSERROR("FCPPR[BIJED]",status,errtxt);
	  } /* if (naxis>0) */
	}
	} /* (FALSE == SDiscoordvar(sds_id)) { */
	
	status = SDendaccess(sds_id);
	if (status != SUCCEED)
	    return(err_msg(WriteFITSFile, FileSystemError));	
	
	/* get the identifier of the next data set  */
	++sds_index;
	sds_id = SDselect(sd_id, sds_index);
     		
    } /* while  (sds_index < numDataset )   */
    
    return(AllOk);

} /* top */



