/* sdsutils.c
 * Assorted utilities for the file conversion package.  
 *
 * 
 */

#include <stdio.h> 
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#define LF 10
#define CR 13
#define FILESIZE 255

/* Package include files */
#include "reformat.h"
#include "types.h"
#include "error.h"
#include "extern.h"


/* Package include files */
#include "fits2sds.h"
#include "cfitsio.h"

/* HDF include files */
#include "hdf.h"
#include "herr.h"

/* Make a new temporary file name.  Result only valid until the routine is
 * called again (statat(ic space).
 */

struct class classTab[] = {
    "SIMPLE",   SIMPLE,
    "BITPIX",   BITPIX,
    "NAXIS",    NAXIS,
    "BSCALE",   BSCALE,
    "BZERO",    BZERO, 
    "DATAMAX",  DATAMAX,
    "DATAMIN",  DATAMIN,
    "BUNIT",    BUNIT,
    "BLANK",    BLANK,
    "EXTNAME",  EXTNAME,
    "EXTVER",   EXTVER,
    "HISTORY",  HISTORY,
    "COMMENT",  COMMENT,

    "BSCALERR",BSCALERR,
    "BZEROERR",BZEROERR,
    "DATATYPE",DATATYPE,
    
    "END",      END
};

#define NCLASSES (sizeof classTab / sizeof(struct class) )

struct fits2sdsAttr fits2sdsAttrTab[]={

    "SIMPLE" , DFNT_UINT8,
    "EXTEND" , DFNT_UINT8,
    /*
    "BSCALE" , DFNT_FLOAT32,
    "BZERO"  , DFNT_FLOAT32,
    "DATAMAX", DFNT_FLOAT32,
    "DATAMIN", DFNT_FLOAT32,
    */
    "PCOUNT",  DFNT_INT32,
    "GCOUNT",  DFNT_INT32,

    "DATE"   , DFNT_UINT8,
    "TITLE"  , DFNT_UINT8,
    "ORIGIN" , DFNT_UINT8,
    "BLOCKED", DFNT_UINT8,
    "AUTHOR" , DFNT_UINT8,
    "REFERENC",DFNT_UINT8,
    "DATE-OBS",DFNT_UINT8,
    "TELESCOP",DFNT_UINT8,
    "INSTRUME",DFNT_UINT8,
    "OBSERVER",DFNT_UINT8,
    "OBJECT"  ,DFNT_UINT8,
    "EQUINOX" ,DFNT_FLOAT32,
    "EPOCH"   ,DFNT_FLOAT32,

    /* BUNIT can this keyword convert to sds' dataset unit? */
    /* "BUNIT"   ,DFNT_UINT8, */
    /* CTYPEn this keyword should be converted to sds' dimension unit  */
    /* "CTYPE"  ,DFNT_UINT8,     
    "CRPIX"  ,DFNT_FLOAT32,
    "CROTA"  ,DFNT_FLOAT32,
    "CRVAL"  ,DFNT_FLOAT32,
    "CDELT"  ,DFNT_FLOAT32,
    */
    /*   
    "COMMENT",DFNT_UINT8,
    "HISTORY",DFNT_UINT8,
    "EXTNAME" ,DFNT_UINT8,
    "EXTLEVEL",DFNT_INT32,
    */
    "XTENSION",DFNT_UINT8,
    "FILENAME",DFNT_UINT8,
    "EXTVER"  ,DFNT_INT32,
    
};

#define NFITS2SDSATTR (sizeof fits2sdsAttrTab / sizeof(struct fits2sdsAttr) )


static char t_file[FILESIZE];
static int file_suffix = 0;
char tempStr[4000];  /* assume maxium 50 rows comments */
char errtxt[FITS_CLEN_ERRMSG];



/* trim left space of the string */

char *ltrim(str)
    char *str;
{
   
    int  i=0;
  
    while ((*str)== ' ')
	str++;
    while ((*str) != '\0') {
	
	tempStr[i] = *str++;
	i++;
    }
    tempStr[i] = '\0';
    return(tempStr);
}
	

/* trim right space of the string */

char *trim(str)
    char *str;
{
   
    int  i=strlen(str);
  
    while ((str[i-1]== ' ') && (i>0))
	--i;

    tempStr[i]='\0';
    --i;
    while (i>=0) {
	
	tempStr[i] = str[i];
	--i;
    }

    return(tempStr);
}


/* find number type based on bitpix  */

int findNumberType(bitpix)

    int bitpix;
{
    int refBitpix;

    refBitpix = bitpix + 100;
    if (refBitpix < 0)
	refBitpix = -refBitpix;
	
    switch(refBitpix) {
    case 108:    /* char  (100+8)  */
	return (DFNT_UINT8);
    case 116:   /* short integer  */
	return (DFNT_INT16);
    case 132:  /* integer  */
	return (DFNT_INT32);
    case 68:  /* float  */
	return(DFNT_FLOAT32);
    case 36:  /* double  */
	return(DFNT_FLOAT64);
    default:
	return(err_msg(Fits2HDF, BadInputFile));		
    }
}




/* check error after make a fits routine call */
void CHECKFITSERROR(routineName, status, errTxt) 
    char routineName[10];
    int status;
    char errTxt[FITS_CLEN_ERRMSG];
{
    if (status != 0){ 
        FCGERR(status,errTxt);
        printf ("%s status = %d : %s\n", routineName, status, errTxt);
	/*	Quit(-1); */
    }
}

/* open FITS file with dersied mode  */

int openFits(fileName, mode)
    char *fileName;
    int mode;
{
    
    int iounit,status=0;
    int bksize,rwstat;

    rwstat = mode;

    /*  get an unused logical unit number */
    FCGIOU(&iounit, &status);
    CHECKFITSERROR("FCGIOU", status, errtxt);

     /* open fits file */
    FCOPEN(iounit, fileName, rwstat, &bksize, &status);
    CHECKFITSERROR("FCOPEN", status, errtxt);

    /* return the result */
    if (status == 0) 
	return(iounit);
    else { 
	printf("Can't open file: %s\n", fileName);
	return(-1);
    }
}

/* open & initialize a new empty FITS file  */

int createFits(fileName, mode)
    char *fileName;
    int mode;
{
    
    int iounit,status=0;
    int bksize,rwstat;

    rwstat = mode;

    /*  get an unused logical unit number */
    FCGIOU(&iounit, &status);
    CHECKFITSERROR("FCGIOU", status, errtxt);

     /* create  fits file */
    bksize = 2800;
    FCINIT(iounit, fileName, bksize, &status);
    CHECKFITSERROR("FCINIT", status, errtxt);

    /* return the result */
    if (status == 0) 
	return(iounit);
    else { 
	printf("Can't create file: %s\n", fileName);
	return(-1);
    }
}
   

/* write attribute to sds */
void writeSdsAttr(id, keyword, numberType, value)
    int32 id, numberType;
    char *keyword;
    char *value;
{
    int charLen ;

    if (strlen((char *)ltrim(keyword))) {  /* make sure keyword is not empty  */
    if (numberType == DFNT_UINT8) {

        strcpy(value,(char *)ltrim(value));
	if (strcmp(substr(value,strlen(value),1),"/")==0)
	  strcpy(value,substr(value,1,strlen(value)-1));
	strcpy(value, (char *)trim(value));

	if ((strcmp(substr(value,1,1),"'")==0) && (strcmp(substr(value,strlen(value),1),"'")==0))
	  strcpy(value,substr(value,2,strlen(value)-2));

	charLen = strlen(value);
	if (charLen)
	  SDsetattr(id, keyword, DFNT_CHAR8, charLen, value); 
    }
    else 
	switch(numberType) {
	case DFNT_INT16:
	{int16 attrVal16;
	attrVal16 = (int16) atoi(value);
	SDsetattr(id, keyword, numberType, 1, &attrVal16);
	}
	break;
	case DFNT_INT32:
	{int32 attrVal32;
	attrVal32 = (int32) atoi(value);
	SDsetattr(id, keyword, numberType, 1, &attrVal32);
	}
	break;	
	case DFNT_FLOAT32:
	{float32 attrFVal32;
	attrFVal32 = (float32) atof(value);
	SDsetattr(id, keyword, numberType, 1, &attrFVal32);
	}
	break;
	case DFNT_FLOAT64:
	{float64 attrFVal64;
	attrFVal64 = (float64) atof(value);
	SDsetattr(id, keyword, numberType, 1, &attrFVal64);
	}
	break;
	}
    } /*  if (strlen(ltrim(keyword)))   */
}


/* write attribute's comment  to sds */
void writeSdsAttrComm(id, keyword, numberType, value)
    int32 id, numberType;
    char *keyword;
    char *value;
{
    int charLen ,tmpLen;
    char commKey[20], commVal[60];
    
    tmpLen = at("/ ",value);
    if (tmpLen>=0) {
      tmpLen += 2;     
      strcpy(commVal,substr(value,tmpLen,50));


      strcpy(commKey,(char *)trim(keyword));
      strcat(commKey,"_COMMENT");
      strcpy(commVal, (char *)ltrim(commVal));
      strcpy(commVal, commVal);
      charLen = strlen(commVal);
      if (charLen) 
	SDsetattr(id, commKey, DFNT_CHAR8, charLen, commVal); 
    }
}


/* get next record's keyword */	
char *getKey(unit, keyno)
    int unit, keyno;
{
   
    char value[24];
    char comment[60];
    int  status = 0;
    char card[80];

    FCGREC(unit, keyno, card, &status);
    strcpy(tempStr, substr(card,1,8));

    if (status == 0) 
	return(tempStr);
    else
	return('\0');
}



/* get fits' keyword attribute in the sds */

int locFitsAttrKey(attrKeyword)
    char attrKeyword[8];
{
    int i,arrayNum;
    int tmpLen;
    arrayNum = NFITS2SDSATTR;
    for (i=0; i<arrayNum; i++) {
	tmpLen = strlen(fits2sdsAttrTab[i].fitsKeyword);
	if (strncmp(attrKeyword, fits2sdsAttrTab[i].fitsKeyword,tmpLen) == 0)
	    return (fits2sdsAttrTab[i].sdsNumberType);
    }
    return(-1);
}

/* return keyword index in the classTab  */

int locateKeywordIndex(priKeyword)
    char priKeyword[8];
{
    int i,arrayNum;
    int tmpLen;
    arrayNum = NCLASSES;
    for (i=0; i< arrayNum; i++) {
	tmpLen = strlen(classTab[i].word);
	if (strncmp((classTab[i].word), priKeyword,tmpLen) == 0)
	    return(classTab[i].value);
    }
    return(-1);
}


/* subroutine of SDsetrange */
void writeSDrange(sdsid, bitpix, max, min)
    int32 sdsid;
    int   bitpix;
    float32 *max,*min;
{
    int32 numtype;
    int16 max16,min16,dataint16[2];
    int32 max32,min32,dataint32[2];
    float64 max64,min64,datafloat64[2];
    float32 datafloat32[2];
    
    numtype = findNumberType(bitpix);
    switch(numtype) {
    case DFNT_INT16:
	max16 = (dataint16[1] = (int16)(*max));
	min16 = (dataint16[0] = (int16)(*min));
	/* SDsetattr(sdsid,"valid_range",numtype,2,&dataint16[0]); */
	SDsetrange(sdsid, &max16, &min16);
	break;

    case DFNT_INT32:
	dataint32[1] = (max32 =(int32)(*max));
	dataint32[0] = (min32 =(int32)(*min));
	/* SDsetrange(sdsid,&max32,&min32); */
	SDsetattr(sdsid,"valid_range",numtype,2,dataint32);
	break;

    case DFNT_FLOAT32:
	datafloat32[1] = *max;
	datafloat32[0] = *min;
	/* SDsetrange(sdsid,max,min); */
	SDsetattr(sdsid,"valid_range",numtype,2,datafloat32);
	break;
    case DFNT_FLOAT64:
	datafloat64[1] = (max64 = (float64)(*max));
	datafloat64[0] = (min64 = (float64)(*min));
	/* SDsetrange(sdsid,&max64,&min64); */
	SDsetattr(sdsid,"valid_range",numtype,2,datafloat64);
	break;
    }
}

/* subroutine of SDsetfillvalue */
void writeSDfillval(sdsid, bitpix, val)
    int32 sdsid;
    int   bitpix;
    float32 val;
{
    int32 numtype;
    int16 val16;
    int32 val32;
    float64 val64;
    
    
    numtype = findNumberType(bitpix);
    switch(numtype) {
    case DFNT_INT16:
	val16 = (int16)val;
	SDsetfillvalue(sdsid, &val16);

	break;

    case DFNT_INT32:
	val32 = (int32)val;
	SDsetfillvalue(sdsid, &val32);

	break;

    }
}


/* get the fits keyword's value type based on value (string) */
int getKeyType(strval)
    char *strval;
{
    int i;
    char tempStr[80];

    strcpy(tempStr,ltrim(strval));    
    if (strcmp(substr(tempStr,strlen(tempStr),1),"/")==0)
      strcpy(tempStr,substr(tempStr,1,strlen(tempStr)-1));
    strcpy(tempStr, trim(tempStr));

    if (strlen(tempStr) == 0)
      return(DFNT_UINT8);
   
    /* fits keyword is String  */
    if (!(strcmp(substr(tempStr,1,1), "'")) && (!strcmp(substr(tempStr,strlen(tempStr),1),"'")))  
	return(DFNT_UINT8);

    /* double precise number --->>> change to float */
    if (((at("E+",tempStr)>=0) || (at("E-",tempStr)>=0)) && (at(".",tempStr)>=0))
	return(DFNT_FLOAT32);

    /* float number */
    if (at(".",tempStr)>=0)
	return(DFNT_FLOAT32);

    /* integer number */
    
    i = (int)atoi(substr(tempStr,1,1));
    if (i)    /* integer */
	return(DFNT_INT32);
    else
	return(DFNT_INT16); /*  0  */
}

	    

/* get BSCALE factor so as to convert float to int , I hope not to lose precision  */
int getBscaleFactor(bitpix,databuffer,datanum)
    int32 bitpix;
    float *databuffer;
    int   datanum;
{
    int32 j,i=1;
    float32 val;
    float32 max, min;

    min = max = *databuffer;

    /* find maxinum & minium  of data  */
    for (j=1;j<=datanum;j++) {
	val = *databuffer++;
	/* val = (val < 0) ? (-val) : val; */
	if (val>max)
	    max = val;
	if (val<min)
	    min = val;
    }
    
    /* range of the dataset  */
    val = max - min;

    if ( val > 100)  /* data distribution is enough for make a image  */
	return(1);

    /* get absolute value of minium  */
    min = (min < 0) ? (-min) : min;
    
    if (min==0)
	/* assume minium is 0.00001 */
	min = 0.00001;

    /* if min < 1  */
    while(1) {

	if (min<1) {
	    i = i * 10;
	    min = min * i;
	}
	else
	    break;
    }

    /* decide factor based on bitpix */
    max = max * i;
    if (bitpix == 32) {  /* 32 bit */
	while(1) {
	    if (max>1000000000)
		i = i/10;
	    else
		break;
	}
    }
    else { /* 16 bit */
	while (1) {
	    if (max > 30000)
		i = i/10;
	    else
		break;
	}
    }

    /* if range of dataset too short and manification factor is one
       then adjust it */
    if ((val<=1) && (i==1)) {
	while (val<=100) {
	    i = i*10;
	    val = val * i;
	}
    }
    
    /* return bscaleFactor */
    if (i > 0)
	return(i);
    else 
	return(1);
}
	    


void findMaxMin8(databuffer,datanum,max,min)
    uint8 *databuffer;
    int   datanum;
    float32 *max,*min;

{
    int32 j;
    uint8 val;
    uint8 max8,min8;
    max8 =min8=*databuffer;

    /* find maxinum & minium  of data  */
    for (j=1;j<=datanum;j++) {
	val = *databuffer++;
	if (val>max8)
	    max8 = val;
	if (val<min8)
	    min8 = val;
    }

    *max = max8;
    *min = min8;

}

void findMaxMin16(databuffer,datanum,max,min)
    int16 *databuffer;
    int   datanum;
    float32 *max,*min;

{
    int32 j;
    int16 val;
    int16 max16,min16;
    max16=min16=*databuffer;

    /* find maxinum & minium  of data  */
    for (j=1;j<=datanum;j++) {
	val = *databuffer++;
	if (val>max16)
	    max16 = val;
	if (val<min16)
	    min16 = val;
    }

    *max = max16;
    *min = min16;

}



void findMaxMin32(databuffer,datanum,max,min)
    int32 *databuffer;
    int   datanum;
    float32 *max,*min;

{
    int32 j;
    int32 val;
    int32 max32,min32;
    max32=min32=*databuffer;

    /* find maxinum & minium  of data  */
    for (j=1;j<=datanum;j++) {
	val = *databuffer++;
	if (val>max32)
	    max32 = val;
	if (val<min32)
	    min32 = val;
    }

    *max = max32;
    *min = min32;

}




/* reverse order of lines in image  */
/* convert fits data to hdf form */
void fdat2cdat_uint8(bitpix,rank,dim,data)
    int32  bitpix, rank;
    int32 *dim;
    uint8 *data;
{
    int   dataNum=1, i, index;
    int   x=0,y=0,z=0,f=0;
    int   datasize ;
    uint8 *sptr;  /* point to one row fits data  */
    uint8  *dataPtr, *ptr;

    
    if (rank>4) 
	InfoMsg("FITS data dimensions is greater than 4 !");
    else {
	for (i=0; i<rank; i++)
	    dataNum *= dim[i];
	
	datasize = (bitpix < 0 ? -bitpix : bitpix)/8 * dataNum;

	dataPtr= ptr = (uint8 *) malloc(dataNum);
	if (rank == 2) {/* two dimension */
	    for (y = dim[0] -1; y>=0; y--) { /* colunm */
		sptr = (uint8 *)&data[y*dim[1]];
		for (x=0; x<dim[1]; x++, ptr++)  /* rows  */
		    *ptr = (uint8) *sptr++;
	    }
	}
	
	if (rank == 3) { /* three dimension data */
	    for (z = 0; z < dim[0]; z++) {
		for (y = dim[1] -1; y>=0; y--) { /* colunm */
		  index = (z*dim[1] + y) * dim[2];
		  sptr = (uint8 *)&data[index];
		  for (x=0; x<dim[2]; x++, ptr++)  /* rows  */
		    *ptr = (uint8) *sptr++;
		}
	    }
	}
	
	
	if (rank == 4) { /* four dimensions data */
	    for (f = 0; f < dim[0]; f++) {
		for (z = 0; z < dim[1]; z++) {
		    for (y = dim[2] -1; y>=0; y--) { /* colunm */
		      index = ((f*dim[1]+z)*dim[2]+y)*dim[3];
		      sptr = (uint8 *)&data[index];
		      for (x=0; x<dim[3]; x++, ptr++)  /* rows  */
			*ptr = (uint8) *sptr++;
		    }
		}	   
	    }
	}
	for (i=0; i<= dataNum; i++)
	    *data++ = (uint8) *dataPtr++;
	
    }
}
    


/* convert fits data to hdf form */
void fdat2cdat_int16(bitpix, rank,dim,data)
    int32  bitpix, rank;
    int32 *dim;
    int16 *data;
{
    int   dataNum=1, i, index;
    int   x=0,y=0,z=0,f=0;
    int   datasize ;
    int16 *sptr;  /* point to one row fits data  */
    int16  *dataPtr, *ptr;

    if (rank>4) 
	InfoMsg("FITS data dimensions is greater than 4 !");
    else {
	for (i=0; i<rank; i++)
	    dataNum *= dim[i];
	
	datasize = (bitpix < 0 ? -bitpix : bitpix)/8 * dataNum;

	dataPtr= ptr = (int16 *) malloc(datasize);
	if (rank == 2) {/* two dimension */
	    for (y = dim[0] -1; y>=0; y--) { /* colunm */
	      index =  y * dim[1];
	      sptr = (int16 *)&data[index];
	      for (x=0; x<dim[1]; x++, ptr++)  /* rows  */
		*ptr = (int16) *sptr++;
	    }
	}
	
	if (rank == 3) { /* three dimension data */
	    for (z = 0; z < dim[0]; z++) {
		for (y = dim[1] -1; y>=0; y--) { /* colunm */
		  index = (z*dim[1] + y) * dim[2];
		  sptr = (int16 *)&data[index];
		  for (x=0; x<dim[2]; x++, ptr++)  /* rows  */
		    *ptr = (int16) *sptr++;
		}
	    }
	}
	
	
	if (rank == 4) { /* four dimensions data */
	    for (f = 0; f < dim[0]; f++) {
		for (z = 0; z < dim[1]; z++) {
		    for (y = dim[2] -1; y>=0; y--) { /* colunm */
		      index = ((f*dim[1]+z)*dim[2]+y)*dim[3];
		      sptr = (int16 *)&data[index];
		      for (x=0; x<dim[3]; x++, ptr++)  /* rows  */
			*ptr = (int16) *sptr++;
		    }
		}	   
	    }
	}
	for (i=0; i<= dataNum; i++)
	    *data++ = (int16) *dataPtr++;
	
   }
}
    




/* convert fits data to hdf form */
void fdat2cdat_int32(bitpix, rank,dim,data)
    int32  bitpix, rank;
    int32 *dim;
    int32 *data;
{
    int   dataNum=1, i, index;
    int   x=0,y=0,z=0,f=0;
    int   datasize ;
    int32 *sptr;  /* point to one row fits data  */

    int32  *dataPtr, *ptr;

    if (rank>4) 
	InfoMsg("FITS data dimensions is greater than 4 !");
    else {
	for (i=0; i<rank; i++)
	    dataNum *= dim[i];
	
	datasize = (bitpix < 0 ? -bitpix : bitpix)/8 * dataNum;

	dataPtr= ptr = (int32 *) malloc(datasize);
	if (rank == 2) {/* two dimension */
	    for (y = dim[0] -1; y>=0; y--) { /* colunm */
		sptr = (int32 *)&data[y*dim[1]];
		for (x=0; x<dim[1]; x++, ptr++)  /* rows  */
		    *ptr = (int32) *sptr++;
	    }
	}
	
	if (rank == 3) { /* three dimension data */
	    for (z = 0; z < dim[0]; z++) {
		for (y = dim[1] -1; y>=0; y--) { /* colunm */
		  index = (z*dim[1] + y) * dim[2];
		  sptr = (int32 *)&data[index];
		  for (x=0; x<dim[2]; x++, ptr++)  /* rows  */
		    *ptr = (int32) *sptr++;
		}
	    }
	}
	
	
	if (rank == 4) { /* four dimensions data */
	    for (f = 0; f < dim[0]; f++) {
		for (z = 0; z < dim[1]; z++) {
		    for (y = dim[2] -1; y>=0; y--) { /* colunm */
		      index = ((f*dim[1]+z)*dim[2]+y)*dim[3];
		      sptr = (int32 *)&data[index];
		      for (x=0; x<dim[3]; x++, ptr++)  /* rows  */
			*ptr = (int32) *sptr++;
		    }
		}	   
	    }
	}
	
	for (i=0; i<= dataNum; i++)
	    *data++ = (int32 ) *dataPtr++;
     }
}
    




/* convert fits data to hdf form */
void fdat2cdat_float32(bitpix, rank,dim,data)
    int32  bitpix, rank;
    int32 *dim;
    float32 *data;
{
    int   dataNum=1, i, index;
    int   x=0,y=0,z=0,f=0;
    int   datasize ;
    float32 *sptr;  /* point to one row fits data  */

    float32  *dataPtr, *ptr;

    if (rank>4) 
	InfoMsg("FITS data dimensions is greater than 4 !");
    else {
	for (i=0; i<rank; i++)
	    dataNum *= dim[i];
	
	datasize = (bitpix < 0 ? -bitpix : bitpix)/8 * dataNum;

	dataPtr= ptr = (float32 *) malloc(datasize);
	if (rank == 2) {/* two dimension */
	    for (y = dim[0] -1; y>=0; y--) { /* colunm */
		sptr = (float32 *)&data[y*dim[1]];
		for (x=0; x<dim[1]; x++, ptr++)  /* rows  */
		    *ptr = (float32) *sptr++;
	    }
	}
	
	if (rank == 3) { /* three dimension data */
	    for (z = 0; z < dim[0]; z++) {
		for (y = dim[1] -1; y>=0; y--) { /* colunm */
		  index = (z*dim[1] + y) * dim[2];
		    sptr = (float32 *)&data[index];
		    for (x=0; x<dim[2]; x++, ptr++)  /* rows  */
			*ptr = (float32) *sptr++;
		}
	    }
	}
	
	
	if (rank == 4) { /* four dimensions data */
	    for (f = 0; f < dim[0]; f++) {
		for (z = 0; z < dim[1]; z++) {
		    for (y = dim[2] -1; y>=0; y--) { /* colunm */
		      index = ((f*dim[1]+z)*dim[2]+y)*dim[3];
		      sptr = (float32 *)&data[index];
		      for (x=0; x<dim[3]; x++, ptr++)  /* rows  */
			*ptr = (float32) *sptr++;
		    }
		}	   
	    }
	}
	for (i=0; i<= dataNum; i++)
	    *data++ = (float32 ) *dataPtr++;
 	
    }
}
    

/* reverse order of lines in image  */
/* convert fits data to hdf form */
void fdat2cdat_float64(bitpix, rank,dim,data)
    int32  bitpix, rank;
    int32 *dim;
    float64 *data;
{
    int   dataNum=1, i, index;
    int   x=0,y=0,z=0,f=0;
    int   datasize ;
    float64 *sptr;  /* point to one row fits data  */

    float64  *dataPtr, *ptr;

    if (rank>4) 
	InfoMsg("FITS data dimensions is greater than 4 !");
    else {
	for (i=0; i<rank; i++)
	    dataNum *= dim[i];
	
	datasize = (bitpix < 0 ? -bitpix : bitpix)/8 * dataNum;

	dataPtr= ptr = (float64 *) malloc(datasize);
	if (rank == 2) {/* two dimension */
	    for (y = dim[0] -1; y>=0; y--) { /* colunm */
		sptr = (float64 *)&data[y*dim[1]];
		for (x=0; x<dim[1]; x++, ptr++)  /* rows  */
		    *ptr = (float64) *sptr++;
	    }
	}
	
	if (rank == 3) { /* three dimension data */
	    for (z = 0; z < dim[0]; z++) {
		for (y = dim[1] -1; y>=0; y--) { /* colunm */
		  index = (z*dim[1] + y) * dim[2];
		  sptr = (float64 *)&data[index];
		  for (x=0; x<dim[2]; x++, ptr++)  /* rows  */
		    *ptr = (float64) *sptr++;
		}
	    }
	}
	
	
	if (rank == 4) { /* four dimensions data */
	    for (f = 0; f < dim[0]; f++) {
		for (z = 0; z < dim[1]; z++) {
		    for (y = dim[2] -1; y>=0; y--) { /* colunm */
		      index = ((f*dim[1]+z)*dim[2]+y)*dim[3];
		      sptr = (float64 *)&data[index];
		      for (x=0; x<dim[3]; x++, ptr++)  /* rows  */
			*ptr = (float64) *sptr++;
		    }
		}	   
	    }
	}

	for (i=0; i<= dataNum; i++)
	    *data++ = (float64) *dataPtr++;
	
    }
}
    

void writeDataStrs(id,labFlag,labVal,unitFlag,unitVal,fmtFlag,fmtVal,coordVal)
int32 id,labFlag,unitFlag,fmtFlag;
char  *labVal,*unitVal,*fmtVal,*coordVal;
{
  if (labFlag) { /* set sds data label  string attribute  */
    if ((unitFlag)&&(fmtFlag))
      SDsetdatastrs(id,labVal ,unitVal ,fmtVal,coordVal);
    else {
      if (unitFlag)
	SDsetdatastrs(id,labVal ,unitVal ,NULL,coordVal);
      else {
	if (fmtFlag)
	  SDsetdatastrs(id,labVal ,NULL,fmtVal,coordVal);
	else   
	  SDsetdatastrs(id,labVal ,NULL,NULL,coordVal);
      }
    }
  }
  else {
    if ((unitFlag)&&(fmtFlag))
      SDsetdatastrs(id,NULL ,unitVal ,fmtVal,coordVal);
    else {
      if (unitFlag)
	SDsetdatastrs(id,NULL ,unitVal ,NULL,coordVal);
      else {
	if (fmtFlag)
	  SDsetdatastrs(id,NULL ,NULL,fmtVal,coordVal);
	else   
	  SDsetdatastrs(id,NULL ,NULL,NULL,coordVal);
      }
    }
  }
}

void writeDimStrs(id,labFlag,labVal,unitFlag,unitVal,fmtFlag,fmtVal)
int32 id,labFlag,unitFlag,fmtFlag;
char  *labVal,*unitVal,*fmtVal;
{
  if (labFlag) { /* set sds data label  string attribute  */
    if ((unitFlag)&&(fmtFlag))
      SDsetdimstrs(id,labVal ,unitVal ,fmtVal);
    else {
      if (unitFlag)
	SDsetdimstrs(id,labVal ,unitVal ,NULL);
      else {
	if (fmtFlag)
	  SDsetdimstrs(id,labVal ,NULL,fmtVal);
	else   
	  SDsetdimstrs(id,labVal ,NULL,NULL);
      }
    }
  }
  else {
    if ((unitFlag)&&(fmtFlag))
      SDsetdimstrs(id,NULL ,unitVal ,fmtVal);
    else {
      if (unitFlag)
	SDsetdimstrs(id,NULL ,unitVal ,NULL);
      else {
	if (fmtFlag)
	  SDsetdimstrs(id,NULL ,NULL,fmtVal);
      }
    }
  }
}


/* convert fits card to generic attributes of sds  */
int writeSdsGenericAttr(sdsid,unit,fitsrecnum)
int32 sdsid,unit,fitsrecnum;
{
  
  int  status, j;
  char card[80],value[24];
  char sdsAttrName[24];
  int  sdsAttrNumType;
  int  count;   /* number of values of SDS attribute */
  
  for (j=1 ; j<=3; j++) {
    status = 0;
    FCGREC(unit,j+fitsrecnum, card, &status);
    CHECKFITSERROR("FCGREC", status, errtxt);

    strcpy(value,substr(card,11,20));

    switch(j) {
    case 1:
      strcpy(sdsAttrName, value);
      break;
    case 2:
      sdsAttrNumType = atoi(value);
      break;
    case 3:
      count = atoi(value);
      break;
    }
  }

  switch(sdsAttrNumType) {
  case DFNT_UINT8: { /* char  */
    uint8 *ptr,*buffer;
    ptr=buffer = (uint8 *)newSpace(count);
    for (j=4;j<count+4; j++) {
      
      status = 0;
      FCGREC(unit,j+fitsrecnum, card, &status);
      CHECKFITSERROR("FCGREC", status, errtxt);

      strcpy(value,substr(card,11,20));

      *ptr++ = (uint8)atoi(value);
    }
    SDsetattr(sdsid, sdsAttrName, DFNT_UINT8, count, buffer);
    
  }
  break;

  case DFNT_INT16: { /* 16-bits integer */
    int16 *ptr,*buffer;
    ptr=buffer = (int16 *)newSpace(count*sizeof(int16));
    for (j=4;j<count+4; j++) {
      
      status = 0;
      FCGREC(unit,j+fitsrecnum, card, &status);
      CHECKFITSERROR("FCGREC", status, errtxt);

      strcpy(value,substr(card,11,20));

      *ptr++ = (int16)atoi(value);
    }
    SDsetattr(sdsid, sdsAttrName, sdsAttrNumType, count, buffer);
  }
  break;
  
  case DFNT_INT32: { /* 32-bits integer */
    int32 *ptr,*buffer;
    ptr=buffer = (int32 *)newSpace(count*sizeof(int32));
    for (j=4;j<count+4; j++) {
      
      status = 0;
      FCGREC(unit,j+fitsrecnum, card, &status);
      CHECKFITSERROR("FCGREC", status, errtxt);

      strcpy(value,substr(card,11,20));

      *ptr++ = (int32)atoi(value);
    }
    SDsetattr(sdsid, sdsAttrName, sdsAttrNumType, count, buffer);
  }
  break;

  case DFNT_FLOAT32: { /* 32-bits float */
    float32 *ptr,*buffer;
    ptr=buffer = (float32 *)newSpace(count*sizeof(float32));
    for (j=4;j<count+4; j++) {
      
      status = 0;
      FCGREC(unit,j+fitsrecnum, card, &status);
      CHECKFITSERROR("FCGREC", status, errtxt);

      strcpy(value,substr(card,11,20));

      *ptr++ = (float32)atof(value);
    }
    SDsetattr(sdsid, sdsAttrName, sdsAttrNumType, count, buffer);
  }
  break;

  case DFNT_FLOAT64: { /* 64-bits double float */
    float64 *ptr,*buffer;
    ptr=buffer = (float64 *)newSpace(count*sizeof(float64));
    for (j=4;j<count+4; j++) {
      
      status = 0;
      FCGREC(unit,j+fitsrecnum, card, &status);
      CHECKFITSERROR("FCGREC", status, errtxt);

      strcpy(value,substr(card,11,20));

      *ptr++ = (float64)atof(value);
    }
    SDsetattr(sdsid, sdsAttrName, sdsAttrNumType, count, buffer);
  }
  break;
  default:
    InfoMsg("Unknown number type of SDS attributes");
  }
  
 

  return(fitsrecnum+3+count);
}

/* the following subroutines is for SDS -> FITS  */
/* get bitpixvalue based on numberType */

int getBitpix(numType)

    int32 numType;
{
    switch(numType) {
    case DFNT_UINT8: case DFNT_UCHAR8:
      return(8);
    case DFNT_INT16: 
      return(16);
    case DFNT_INT32:
      return(32);
    case DFNT_FLOAT32:
      return(-32);
    case DFNT_FLOAT64:
      return(-64);
    case DFNT_INT8:  case DFNT_CHAR8:
      return(8);
    case DFNT_UINT16:
      return(16);
    case DFNT_UINT32:
      return(32);
    default:
	return(err_msg(WriteFITSFile, ImageTooComplex));		
    }
}


/* write info into sdsAttrRec  */
int writeSdsAttrRec(sdsAttrSet, num, keyword, numType, buf, bufCount)
    sdsAttr *sdsAttrSet;
    int num, numType, bufCount;
    char *buf;
    char keyword[24];
{
    int i, n;
    strcpy(sdsAttrSet[num].keyword, keyword);
    sdsAttrSet[num].dataType = numType;
    switch(numType) {
    case  DFNT_CHAR8: /* char */
	buf[bufCount] = '\0';
	/* allocate space for sdsAttrSet[num].cKeywordVal */
	sdsAttrSet[num].cKeywordVal =  (char *)malloc(bufCount);
	if (sdsAttrSet[num].cKeywordVal == NULL)
	    return(err_msg(WriteFITSFile, NoMemory));
	strcpy(sdsAttrSet[num].cKeywordVal, buf);
	sdsAttrSet[num].dataType = DFNT_UINT8;

	break;

    case DFNT_INT16:  /* short */
    {
	/* convert raw data to native data form */
	uint16 *sbuf = (uint16 *)buf;
	/*
	n = bufCount; 
	sbuf += n-1;     
	buf += (n-1)*2;  
	for (i=0; i<n; i++, buf-=2, sbuf--) {
	    *sbuf = (((int)*buf)<<8) | (int)(buf[1]);

	    sdsAttrSet[num].sKeywordVal = (int16)*sbuf;
	} */

	sdsAttrSet[num].sKeywordVal = (int16)*sbuf;
    }
	break;

    case DFNT_INT32: /* int */
    {
	/* convert raw data to native data form  */
	uint32 *ibuf = (uint32 *)buf;
	/*
	n = bufCount;
	ibuf += n-1;     
	buf += (n-1)*4;  
	for (i=0; i<n; i++, buf-=4, ibuf--) {
	    *ibuf = (((int)buf[0])<<24)|(((int)buf[1])<<16)|(((int)buf[2])<<8)|(int)(buf[3]);

	sdsAttrSet[num].iKeywordVal = (int32)*ibuf;
	} */
	sdsAttrSet[num].iKeywordVal = (int32)*ibuf;

    }
	break;

    case DFNT_FLOAT32: /* float */
    {
	/* convert raw data to native data */
	int j,k,expo;
	static float32 *exps=NULL;
	float32 *fbuf = (float32 *)buf;
	/* 
	n = bufCount;
	if (exps == NULL) {
	    exps = (float32 *) malloc(256 * sizeof(float32));
	    if (exps == NULL)
		return(err_msg(WriteFITSFile, NoMemory));
	    exps[150] = 1.;
	    for (i=151; i<256; i++)
		exps[i] = 2.*exps[i-1];
	    for (i=149; i>=0; i--)
		exps[i] = 0.5*exps[i+1];
	}
	for (i=0; i<n; i++,buf += 4) {
	    k = (int)*buf;
	    j=((int)buf[1]<<16)|((int)buf[2]<<8)|((int)buf[3]);
	    expo = ((k & 127)<< 1) | (j >> 23);
	    if ((expo | j) == 0) *(float32 *)buf = 0.;
	    else *(float32 *)buf = exps[expo]*(float32)(j|0x800000);
	    if (k&128)  
		*(float32 *)buf = - *(float32 *)buf;    
	    sdsAttrSet[num].fKeywordVal = (float32)*fbuf;
	}
	
	*/
	sdsAttrSet[num].fKeywordVal = (float32)*fbuf;
    }
	break;

    case DFNT_FLOAT64: /* double float */
    {
	/* convert raw data to native data */
	float64 *fbuf = (float64 *)buf;
	sdsAttrSet[num].fKeywordVal = (float32)*fbuf;
    }
	break;
	
    default:
	InfoMsg("SDS attributes is too complex! ");
    }
}

/* write sds generic attributes info into sdsAttrRec  */

int convSdsGenericAttr(sdsAttrSet, num, keyword, numType, buf, bufCount)
    sdsAttr *sdsAttrSet;
    int num, numType, bufCount;
    char *buf;
    char keyword[24];
{
    int i, n, j=num;
    char comment[80];
    char* spac();

    uint8 *tmpChar  = (uint8 *)buf;
    int16 *tmpInt16 = (int16 *)buf;
    int32 *tmpInt32 = (int32 *)buf;
    float32 *tmpFloat32 = (float32 *)buf;
    float64 *tmpFloat64 = (float64 *)buf;

    /* write tag for sds Attributes  */
    strcpy(sdsAttrSet[j].keyword, "COMMENT");
    sdsAttrSet[j].dataType = DFNT_UINT8;
    /* allocate space for sdsAttrSet[num].cKeywordVal */
    sdsAttrSet[j].cKeywordVal =  (char *)malloc(80);
    strcpy(sdsAttrSet[j].cKeywordVal, "SDS attributes");
    ++j;

    
    /* write name  for sds Attributes  */
    strcpy(sdsAttrSet[j].keyword, "blank");
    sdsAttrSet[j].dataType = DFNT_UINT8;
    /* allocate space for sdsAttrSet[num].cKeywordVal */
    sdsAttrSet[j].cKeywordVal =  (char *)malloc(80);
    strcpy(comment,"  ");
    strcat(comment,keyword);
    n=strlen(keyword);
    strcat(comment, spac(20-n));
    strcat(comment, " / Name of SDS attributes");
    strcpy(sdsAttrSet[j].cKeywordVal, comment);
    ++j;

    /* write numberType  for sds Attributes  */
    strcpy(sdsAttrSet[j].keyword, "blank");
    sdsAttrSet[j].dataType = DFNT_UINT8;
    /* allocate space for sdsAttrSet[num].cKeywordVal */
    sdsAttrSet[j].cKeywordVal =  (char *)malloc(80);

    /* strcpy(comment,num2str(numType)); */
    sprintf(comment,"%22d", numType);
    strcat(comment, " / Number type  of SDS attributes");
    strcpy(sdsAttrSet[j].cKeywordVal, comment);
    ++j;



    /* write number of value  for sds Attributes  */
    strcpy(sdsAttrSet[j].keyword, "blank");
    sdsAttrSet[j].dataType = DFNT_UINT8;
    /* allocate space for sdsAttrSet[num].cKeywordVal */
    sdsAttrSet[j].cKeywordVal =  (char *)malloc(80);

    /* strcpy(comment,num2str(bufCount)); */
    sprintf(comment,"%22d",bufCount);
    strcat(comment, " / Number of values  of SDS attributes");
    strcpy(sdsAttrSet[j].cKeywordVal, comment);
    ++j;

    
    for (i=1; i<=bufCount; i++) {
      
      /* write number of value  for sds Attributes  */
      strcpy(sdsAttrSet[j].keyword, "blank");
      sdsAttrSet[j].dataType = DFNT_UINT8;
      /* allocate space for sdsAttrSet[num].cKeywordVal */
      sdsAttrSet[j].cKeywordVal =  (char *)malloc(80);
      switch(numType) {
      case DFNT_UINT8: 
	sprintf(comment,"%22d",(*tmpChar++));
	break;
      case DFNT_INT16:
	sprintf(comment,"%22d",(*tmpInt16++));
	break;
      case DFNT_INT32: 
	sprintf(comment,"%22d",(*tmpInt32++));
	break;
      case DFNT_FLOAT32: 
	sprintf(comment,"%22.6f",(*tmpFloat32++));
	break;
      case DFNT_FLOAT64: 
	sprintf(comment,"%22.6f",(*tmpFloat64++));
	break;  
      }
      strcat(comment," / Value ");
      strcat(comment,num2str(i));
      strcat(comment," of SDS attribute ");
      strcpy(sdsAttrSet[j].cKeywordVal, comment);
      ++j;
    }

    return(j);
}


/* initializing the data array */
void initArray(attrRec,count)
    fitsHead *attrRec;
    int     count;
{
    int i;
    for (i=0;i<=count+10;i++) {
      strcpy(attrRec[i].keyword,    "");
      strcpy(attrRec[i].cKeywordVal,"");
      attrRec[i].sKeywordVal=0;
      attrRec[i].iKeywordVal=0;
      attrRec[i].fKeywordVal=0.0;
      attrRec[i].dataType   =0;
      strcpy(attrRec[i].keywordComm,"");

    }

}



/* lokkup keyword in fitsHeadRec */
int lookupKeyword(attrRec, key, count)
    fitsHead *attrRec;
    char    *key;
    int     count;
{
    int i;
    for (i=0;i<count;i++) 
	if (strncmp(attrRec[i].keyword, key, strlen(key)) == 0)
	    return(i);
    return(-1);
}

/* delete the file */
void deleteFile(fileName)
char *fileName;
{
  int unit,blockSize;
  int status=0;
  char *errtxt[FITS_CLEN_ERRMSG];

  /* get an unused Logical Number to use to open the FITS file */
  FCGIOU(&unit,&status);
  CHECKFITSERROR("FCGIOU",status,errtxt);
  
  /* Try to open the file to see if it exists */
  FCOPEN(unit,fileName,1,&blockSize,&status); 

  if (status == 0) {
    /* delete this file */
    FCDELT(unit,&status);
    CHECKFITSERROR("FCDELT",status,errtxt);
  }
  else {
      /* file doesn't exist, RESET status to zero and clear errors */
      status = 0;
      FCCMSG;
  }
  /* free unit number */
  FCFIOU(unit,&status);
}
  


/* write fits file header */

void writeFitsHead(unit, fitsCard, fitsKeyNumber,sdsNumType)
int      unit;
fitsHead *fitsCard;
int32    sdsNumType,  fitsKeyNumber;
{
  int  status=0;
  int  bitpix, naxis, naxes[99], pcount, gcount;
  int  i,j,loc,num=fitsKeyNumber;
  int  simple,extend,trueFlag;
  char keyword[9];
  char value[20],comment[80],blankComm[80];
  char errtxt[FITS_CLEN_ERRMSG];

  float32 dataMax, dataMin, bscaleVal=1.0,bzeroVal=0.0,fitsBzeroVal=0.0;
     
  int  datamaxFlag=0, dataminFlag=0;
  int  bscaleFlag =0, bzeroFlag  =0;

  /* modify the keyword BSCALE & BZERO  */

  loc = lookupKeyword(fitsCard,"BSCALE",num);
  if (loc >= 0) 
    bscaleVal = fitsCard[loc].fKeywordVal;
  
  loc = lookupKeyword(fitsCard,"BZERO",num);
  if (loc >= 0) {
    bzeroVal = fitsCard[loc].fKeywordVal;
    fitsBzeroVal = -(bscaleVal)*bzeroVal; /* convert to fits bzero value */
    switch(sdsNumType) {
    case DFNT_INT8: case DFNT_CHAR8:      
      fitsBzeroVal +=  -bscaleVal*128;  /* 2**7 */
      break;
    case DFNT_UINT16: 
      fitsBzeroVal +=  bscaleVal*32768;  /* 2**15 */
      break;
    case DFNT_UINT32:
      fitsBzeroVal +=  bscaleVal*2147483648.0;  /* 2**15 */
    }
    fitsCard[loc].fKeywordVal = fitsBzeroVal;
  }
  else {  /* add kerword BZERO if other sdsDataType is defined */
    
    if ((sdsNumType==DFNT_INT8)||(sdsNumType==DFNT_UINT16)||(sdsNumType==DFNT_CHAR8)||(sdsNumType==DFNT_UINT32)) {
      ++num;
      strcpy(fitsCard[num-1].keyword,"BZERO");
      fitsCard[num-1].dataType = DFNT_FLOAT32;
      switch(sdsNumType) {
      case DFNT_INT8: case DFNT_CHAR8:      
	fitsBzeroVal =  -bscaleVal*128;  /* 2**7 */
	break;
      case DFNT_UINT16: 
	fitsBzeroVal =  bscaleVal*32768;  /* 2**15 */
	break;
      case DFNT_UINT32:
	fitsBzeroVal =  bscaleVal*2147483648.0;  /* 2**15 */
      }
      fitsCard[num-1].fKeywordVal = fitsBzeroVal;
    }
  }


  /* the keyword DATAMAX & DATAMIN  */

  loc = lookupKeyword(fitsCard,"DATAMAX",num);
  if (loc >= 0) {
    switch(fitsCard[loc].dataType) {
    case DFNT_INT16:
      dataMax = (float32)(fitsCard[loc].sKeywordVal);
      break;
    case DFNT_INT32:     
      dataMax = (float32)(fitsCard[loc].iKeywordVal);
      break;
    case DFNT_FLOAT32:
      dataMax = (float32)(fitsCard[loc].fKeywordVal);
      break;
    }
    switch(sdsNumType) { /* modified datamax .. */
    case DFNT_INT8: case DFNT_CHAR8:      
      dataMax += 128;
      break;
    case DFNT_UINT16: 
      dataMax -= 32768;
      break;
    case DFNT_UINT32:
      dataMax -= 2147483648.0;  /* 2**15 */
    }

    dataMax = (dataMax*bscaleVal)+fitsBzeroVal; /* convert it to fits keyword value */


    /* change dataType to float32 for fits keyword DATAMAX */
    /* 
       switch(fitsCard[loc].dataType) {
       case DFNT_INT16:
       fitsCard[loc].sKeywordVal = (int16)dataMax;
       break;
       case DFNT_INT32:     
       fitsCard[loc].iKeywordVal = (int32)dataMax;
       break;
       case DFNT_FLOAT32:
       fitsCard[loc].fKeywordVal = dataMax;
       break;
       }
       */
    fitsCard[loc].dataType    = DFNT_FLOAT32;
    fitsCard[loc].fKeywordVal = dataMax;
  }
 

  /* DATAMIN  */
  loc = lookupKeyword(fitsCard,"DATAMIN",num);
  if (loc >= 0) {
    switch(fitsCard[loc].dataType) {
    case DFNT_INT16:
      dataMin = (int16)  (fitsCard[loc].sKeywordVal);
      break;
    case DFNT_INT32:     
      dataMin = (float32)(fitsCard[loc].iKeywordVal);
      break;
    case DFNT_FLOAT32:
      dataMin = (float32)(fitsCard[loc].fKeywordVal);
      break;
    }
    switch(sdsNumType) { /* modified datamin
 .. */
    case DFNT_INT8: case DFNT_CHAR8:     
      dataMin += 128;
      break;
    case DFNT_UINT16: 
      dataMin -= 32768;
      break;
    case DFNT_UINT32:
      dataMin -= 2147483648.0;  /* 2**15 */
    }
    dataMin = (dataMin*bscaleVal)+fitsBzeroVal; /* convert it to fits keyword value */
    /*
      switch(fitsCard[loc].dataType) {
      case DFNT_INT16:
      fitsCard[loc].sKeywordVal = (int16)dataMin;
      break;
      case DFNT_INT32:     
      fitsCard[loc].iKeywordVal = (int32)dataMin;
      break;
      case DFNT_FLOAT32:
      fitsCard[loc].fKeywordVal = dataMin;
      break;
      }
      */
    fitsCard[loc].dataType    = DFNT_FLOAT32;
    fitsCard[loc].fKeywordVal = dataMin;
  }
  

  /* if XTENSION is exist then create new PHDU */
  loc = lookupKeyword(fitsCard,"XTENSION",num);
  if (loc >= 0) {
    /* create a new empty HDU at end of fits file  */
    FCCRHD(unit,&status);
    CHECKFITSERROR("FCCRHD",status,errtxt);
  }



  for (i=0; i<num; i++) {
    status=0;
    strcpy(keyword, fitsCard[i].keyword);
    strcpy(comment, fitsCard[i].keywordComm);
    
    if (strncmp(keyword, "blank", 5) == 0) { /* blank */
      strcpy(blankComm,"        ");
      strcat(blankComm,comment);
      FCPREC(unit,blankComm,&status);
      CHECKFITSERROR("FCPREC",status,errtxt);
      continue;

    }


    if (strncmp(keyword ,"COMMENT",7) == 0) {/* comment */
      strcpy(comment,ltrim(comment));
      FCPCOM(unit,comment,&status);
      CHECKFITSERROR("FCPCOM",status,errtxt);
    }
    else { /* comment */
      if (strncmp(keyword ,"HISTORY",7) == 0) {/* HISTORY */
	strcpy(comment,ltrim(comment));
	FCPHIS(unit,comment,&status);
	CHECKFITSERROR("FCPHIS",status,errtxt);
      }
      else {

	switch(fitsCard[i].dataType) { 
	case DFNT_UINT8:  /* such as SIMPLE ... */
	  FCPKYS(unit, keyword, fitsCard[i].cKeywordVal,comment, &status);
	  break;
	case DFNT_INT16:  /*  ... */
	  FCPKYJ(unit, keyword, fitsCard[i].sKeywordVal,comment, &status);
	  break;
	case DFNT_INT32:  /* such as BITPIX ... */
	  FCPKYJ(unit, keyword, fitsCard[i].iKeywordVal,comment, &status);
	  break;
	case DFNT_FLOAT32: { /* such as DATAMIN ... */

	  float32 tmp=0;
	  tmp = fitsCard[i].fKeywordVal;
	  tmp = (tmp>0)?tmp:(-tmp);

	  /* write fits keyword value based on the value */
	  if (tmp>1.0e+10) /* 1,000,000 */
	    FCPKYE(unit, keyword, fitsCard[i].fKeywordVal,2,comment, &status);
	  else {
	    if (tmp>100) /*  1,000,000 ~ 100 */
	      FCPKYF(unit,keyword,fitsCard[i].fKeywordVal,3,comment,&status);
	    else {
	      if (tmp>10) /*  100 ~ 10 */
		FCPKYF(unit,keyword,fitsCard[i].fKeywordVal,4,comment,&status);
	      else { 
		if ((tmp<1.0e-5) && (tmp != 0.0))  
		  FCPKYE(unit,keyword,fitsCard[i].fKeywordVal,5,comment,&status);
		else {
		  if ((tmp*1000 - (int)(tmp*1000)) < 1.0e-2) /* decimals is 3 */
		    FCPKYF(unit,keyword,fitsCard[i].fKeywordVal,3,comment,&status);
		  else
		    FCPKYE(unit,keyword,fitsCard[i].fKeywordVal,5,comment,&status);
		}
	      }
	    }
	  }
	}   
	break;

	case DFNT_FLOAT64:  /* such as BSCALE & BZERO ... */
	  FCPKYE(unit, keyword, fitsCard[i].fKeywordVal,5,comment, &status);
	  break;

	} /* switch (  ... */
	CHECKFITSERROR("FCPKY[SJFE]",status,errtxt);

      } /* else */
    } /* comment */
  } /* for ... */

  FCPHIS(unit, "This extension converted to FITS from HDF SDS object by reformat", &status);
 
  FCPREC(unit,"END",&status);

  /* modify the keyword SIMPLE & EXTEND  */

  loc = lookupKeyword(fitsCard,"SIMPLE",num);
  if (loc >= 0) {
    strcpy(value,fitsCard[loc].cKeywordVal);
    strcpy(comment,fitsCard[loc].keywordComm);
    status = 0;
    if (strcmp(value,"T")== 0) {
      trueFlag = 1;
      FCMKYL(unit,"SIMPLE",trueFlag,comment,&status);
    }
    else {
      trueFlag = 0;
      FCMKYL(unit,"SIMPLE",trueFlag,comment,&status);
    }
    CHECKFITSERROR("FCMKYL",status,errtxt);
  }
  /* EXTEND */

  loc = lookupKeyword(fitsCard,"EXTEND",num);
  if (loc >= 0) {
    strcpy(value,fitsCard[loc].cKeywordVal);
    strcpy(comment,fitsCard[loc].keywordComm);
    status = 0;
    if (strcmp(value,"T")== 0) {
      trueFlag = 1;
      FCMKYL(unit,"EXTEND",trueFlag,comment,&status);
    }
    else {
      trueFlag = 0;
      FCMKYL(unit,"EXTEND",trueFlag,"&",&status);
    }
    CHECKFITSERROR("FCMKYL",status,errtxt);
  }

  /* BLOCKED */

  loc = lookupKeyword(fitsCard,"BLOCKED",num);
  if (loc >= 0) {
    strcpy(value,fitsCard[loc].cKeywordVal);
    strcpy(comment,fitsCard[loc].keywordComm);
    status = 0;
    if (strcmp(value,"T")== 0) {
      trueFlag = 1;
      FCMKYL(unit,"BLOCKED",trueFlag,comment,&status);
    }
    else {
      trueFlag = 0;
      FCMKYL(unit,"BLOCKED",trueFlag,"&",&status);
    }
    CHECKFITSERROR("FCMKYL",status,errtxt);
  }

 
	   
  FCRDEF(unit,&status);
  CHECKFITSERROR("FCRDEF",status,errtxt);
	  
}
