/* utils.c
 * Assorted utilities for the file conversion package.  
 *
 * Peter Webb, Summer 1990
 */

#include <stdio.h> 
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#define LF 10
#define CR 13

/* Package include files */
#include "reformat.h"
#include "types.h"
#include "error.h"
#include "extern.h"

/* HDF include files */
#include "hdf.h"
#include "herr.h"
static int file_suffix = 0;
static char t_file[FILE_SIZE];
char tmpStr[4000];

/* Make a new temporary file name.  Result only valid until the routine is
 * called again (static space).
 */

char *NewTempFile()
{
  sprintf(t_file, "rfmt_%d.%d", getpid(), file_suffix++);
  return(t_file);
}

/* Allocate and initialize a CvtBlkPtr structure */

CvtBlkPtr NewCvtBlk()
{
  CvtBlkPtr new;

  new = (CvtBlkPtr) malloc(sizeof(CvtBlk));
  if (new == NULL) return(NULL);
  NEXT_BLK(new) = NULL;
  FROM_FILE(new) = NULL;
  TO_FILE(new) = NULL;
  FROM_TYPE(new) = none;
  TO_TYPE(new) = none;
  
  return(new);
}

/* Print out the contents of a FileInfo structure */

void PrintFileInfoStruct(info)
     FileInfo *info;
{
  printf("Info record\n");
  printf(" Name = %s\n", info->name);
  printf(" Dimension = %d\n", info->dimension);
  printf(" Height = %d\n", info->height);
  printf(" Width = %d\n", info->width);
  printf(" Image ptr = %d\n", info->image);
  printf(" Value bitmask = %d\n", info->values);
}

/* Allocate and initialize a StrList structure */

StrListPtr NewStrListElt(str)
  char *str;
{
  StrListPtr new;

  new = (StrListPtr) malloc(sizeof(StrList));
  if (new == NULL) return(NULL);
  if (str != NULL)
    LIST_STR(new) = strsave(str);
  else LIST_STR(new) = NULL;
  NEXT_STR(new) = NULL;
  
  return(new);
}

/* Print an informational message */

void InfoMsg(str)
     char *str;
{
#ifdef XGUI
  WriteOutputMsg(str);
#else
  printf("%s\n", str);
#endif
}

/* Free a list of StrListPtr structures */

ErrorCode FreeStrList(list)
  StrListPtr list;
{
  StrListPtr cur;

  if (list == NULL) return(err_msg(FreeStringList, NullPtr));
  cur = list;
  while (cur != NULL)
    {
      cur = NEXT_STR(list);
      free(LIST_STR(list));
      free(list);
      list = cur;
    }
  return(AllOk);
}

/* Free an array of character strings.  Must be NULL-terminated. */

ErrorCode FreeStrArray(array)
  char *array[];
{
  char *ptr;
  int i = 0;

  if (array == NULL) return(err_msg(FreeStringArray, NullPtr));

  ptr = array[0];
  while (ptr != NULL)
    {
      free(ptr);
      ptr = array[++i];
    }
  return(AllOk);
}

/* Compute x ^ y */

long power (x, y)
  register long x, y;
{
  register long result = 1, i = 0;

  for (i = 0; i < y; i++) result *= x;
  return (result);
}

/* Does a given file exist? */

int FileExists(name)
     char *name;
{
  FILE *fp;

/* Check to see if the file already exists */

  fp = fopen(name, "r");
  if (fp == NULL) return(FALSE);
  return(TRUE);
}

void FileOverWriteMsg(buf, file)
     char *buf, *file;
{
  sprintf(buf, "%s\nalready exists.  Overwrite (y/n)? ", file);
}

/* Confirm overwrite */

int OverWrite(file)
  char *file;
{
  char c = (char)NULL;

/* Check to see if the file already exists */

  if (!FileExists(file)) return(TRUE);

/* Simple dialog with user */

  FileOverWriteMsg(ScratchBuf, file);

  printf("%s", ScratchBuf);
  
  while (c != 'Y' &&c != 'y' &&c != 'n' &&c != 'N' && c != '\n')
    c = getchar();
  if (c == 'Y' || c == 'y') return(TRUE);
  return(FALSE);
}

ErrorCode FullInputPath(name, buf)
  char *name, *buf;
{
  if (name == NULL || buf == NULL)
    return(err_msg(FullInputPathname, NullPtr));

/* Get the current string from the input directory box, and append the
 * file name to it.
 */

  sprintf(buf, "%s/%s", InputDirectory, name);
  return(AllOk);
}

ErrorCode FullOutputPath(name, buf)
  char *name, *buf;
{
  if (name == NULL || buf == NULL)
    return(err_msg(FullOutputPathname, NullPtr));

/* Get the current string from the input directory box, and append the
 * file name to it.
 */

  sprintf(buf, "%s/%s", OutputDirectory, name);
  return(AllOk);
}

 

/* get substr from the string */

char *substr( str,start, stop)
    char *str;
    int  start, stop;
{
    int x ;
    int i = 0;
    if (strlen(str) < start)
        { tmpStr[0] = '\0';
          return(tmpStr);
        }
    {
      for(x=start-1;(stop>i)  ;x++,i++){
          if (strlen(str) < start+i)
              break;
          tmpStr[i] = str[x];
      }
    }
    tmpStr[i] = '\0';
    return(tmpStr);
}
/* convert every character in the string to upper character  */

char* upper(str)
    char *str;
{
    int i=0;

    while ((tmpStr[i] = toupper(str[i])) != '\0')
      i++;
    return(tmpStr);

}

/* look at str1 to see if it is exist in str2   */

int at(str1,str2)
    char *str1, *str2;
{
    int  i=0, len;
    len = strlen(str2);
    if (!str1)
        return(-1);

    while (strncmp(str2,str1,strlen(str1)) != 0 )
        { *str2++;
          i++;
          if (len < i)
              return(-1);
        }

    return(i);
}


/* generate blank string */
char* spac(n)
int n;
{
  int i;
  

  if ((n>100) || (n<=0)) {
    tmpStr[0]='\0';
    return tmpStr;
  }
  for  (i=0; i<n; i++)
    tmpStr[i]=' ';
  tmpStr[n]='\0';
     
  return tmpStr;
}

/*
 * newSpace
 *
 * Allocate a space with little wastage
 *
 */

#ifdef PROTOTYPE
char *newSpace(int32 size)
#else
char *newSpace(size)
    int32 size;
#endif /* PROTOTYPE */
{
    static int32 oldSize = 0;        /* must be static */
    static char *oldSpace = NULL; /* must be static */

    if (size >= oldSize) {
        if (oldSpace != NULL) HDfreespace(oldSpace);
        if ((oldSpace = (char *) HDgetspace((uint32) size)) == NULL) {
            puts("Out of memory. Abort.");
            exit(1);
        }
        oldSize = size;
    }

    return oldSpace;
}

char *num2str(num)
    int num;
{
    if (num < 0)
        num = -num;
    if (num <10)
        sprintf(tmpStr,"%1d",num);
    else
        sprintf(tmpStr,"%2d",num);
    return(tmpStr);
}

/***********************************/
void FatalError (identifier)
      char *identifier;
{
  fprintf(stderr, " %s\n", identifier);
  exit(-1);
}


/***************************************************/
void xvbcopy(src, dst, len)
     char *src, *dst;
     int   len;
{
  /* Modern OS's (Solaris, etc.) frown upon the use of bcopy(),
   * and only want you to use memcpy().  However, memcpy() is broken,
   * in the sense that it doesn't properly handle overlapped regions
   * of memory.  This routine does, and also has its arguments in
   * the same order as bcopy() did, without using bcopy().
   */

  /* determine if the regions overlap
   *
   * 3 cases:  src=dst, src<dst, src>dst
   *
   * if src=dst, they overlap completely, but nothing needs to be moved
   * if src<dst and src+len>dst then they overlap
   * if src>dst and src<dst+len then they overlap
   */

  if (src==dst || len<=0) return;    /* nothin' to do */
  
  if (src<dst && src+len>dst) {  /* do a backward copy */
    src = src + len - 1;
    dst = dst + len - 1;
    for ( ; len>0; len--, src--, dst--) *dst = *src;
  }

  else {  /* they either overlap (src>dst) or they don't overlap */
    /* do a forward copy */
    for ( ; len>0; len--, src++, dst++) *dst = *src;
  }
}
    

/***************************************************/
int xvbcmp (s1, s2, len)
     char *s1, *s2;
     int   len;
{
  for ( ; len>0; len--, s1++, s2++) {
    if      (*s1 < *s2) return -1;
    else if (*s1 > *s2) return 1;
  }
  return 0;
}

/***************************************************/
void xvbzero(s, len)
     char *s;
     int   len;
{
  for ( ; len>0; len--) *s++ = 0;
}

/***************************************************/

void Quit(i)

int i;
{ 
    exit(i);
}


