
/* write_fits.c
 * Write data out to a FITS file.
 *
 */

#include <stdio.h>

/* Package include files */

#include "reformat.h"
#include "types.h"
#include "extern.h"
#include "error.h"

typedef unsigned char uint8;

/* #define intensity(color) (unsigned int) ((unsigned int) ((color).red*27+(color).green*150+(color).blue*29) >>8) */

 
/* MONO returns total intensity of r,g,b triple (i = .33R + .5G + .17B) */
#define MONO(rd,gn,bl) ( ((int)(rd)*11 + (int)(gn)*16 + (int)(bl)*5) >> 5)

#define BLOCKSIZE 2880

/* Write out a FITS file */

ErrorCode WriteFITS(info)
  FileInfo *info;
{
  FILE *fp;
  char
    buffer[81],
    *fits_header;

  register int
    i,nend,np,j;

  uint8  rgb[256], *ptr;
  register uint8  *p;
  unsigned int red[256], green[256], blue[256];



  if (info == NULL) return(err_msg(WriteFITSFile, NullPtr));

  /*
    Open output image file.
  */
  fp = fopen(info->name,"w");
  if (fp == NULL) 
      return(err_msg(WriteFITSFile, OutputOpenFailed));

  /*  Allocate image header.*/

  fits_header=(char *) malloc(2880*sizeof(uint8));
  if (fits_header == (char *) NULL)
      return(err_msg(WriteFITSFile, NoMemory));

  /*
    Initialize image header.
  */
  for (i=0; i < 2880; i++)
    fits_header[i]=' ';
  (void) strcpy(buffer,"SIMPLE  =                    T");
  (void) strncpy(fits_header+0,buffer,strlen(buffer));
  (void) strcpy(buffer,"BITPIX  =                    8");
  (void) strncpy(fits_header+80,buffer,strlen(buffer));
  (void) strcpy(buffer,"NAXIS   =                    2");
  (void) strncpy(fits_header+160,buffer,strlen(buffer));
  (void) sprintf(buffer,"NAXIS1  =           %10u",info->width);
  (void) strncpy(fits_header+240,buffer,strlen(buffer));
  (void) sprintf(buffer,"NAXIS2  =           %10u",info->height);
  (void) strncpy(fits_header+320,buffer,strlen(buffer));
  (void) strcpy(buffer,"HISTORY Created by WriteFITS");
  (void) strncpy(fits_header+400,buffer,strlen(buffer));
  (void) strcpy(buffer,"END");
  (void) strncpy(fits_header+480,buffer,strlen(buffer));
  (void) fwrite((char *) fits_header,1,2880,fp);
  free((char *) fits_header);
  /*
    Convert image to fits scale PseudoColor class.
  */
  /*  fits_pixels=(uint8 *)
    malloc(info->height*info->width*sizeof(uint8));
  if (fits_pixels == (uint8 *) NULL)
      return(err_msg(WriteFITSFile, NoMemory)); */
  
  p = info->image;
  
  for (i=0; i<MAX_PAL/3; i++){
      red[i]  =info->palette[3*i  ];
      green[i]=info->palette[3*i+1];
      blue[i] =info->palette[3*i+2];
  }
  for (i=0; i<MAX_PAL/3; i++)
      rgb[i] = MONO(red[i],green[i],blue[i]);
  
  /*  (void) fwrite(info->image ,(int) (info->height*info->width),1, fp);  */
  for (i=info->height-1; i >= 0; i--) {     /* flip line ordering when writing out */
      ptr = &p[i*info->width];
      for (j=0; j < info->width; j++, ptr++) putc(rgb[*ptr], fp);
  }

  /* nend is the number of padding characters at the end of the last block */
  np = info->width * info->height;
  nend = ((np+BLOCKSIZE-1)/BLOCKSIZE)*BLOCKSIZE - np;
  if (nend) for (i=0; i<nend; i++) putc('\0', fp);
  
  /*  free((uint8 *) fits_pixels); */
  fclose(fp);
  return(AllOk);
}






