/* write_sunrast.c
 * Write data out to a Sun raster file.
 * 
 */

/* Package include files */

#include "reformat.h"
#include "types.h"
#include "extern.h"
#include "error.h"

#include "rasterfile.h"     /* sun raster image format (take it fro SUN)  */


/*  HDF include file     */
#include "hdfi.h"

#define CVT(x) (((x) * 255) / ((1L<<16)-1))

ErrorCode HdfToPixrect(info, header, colormap, pix)
FileInfo      *info;
SunRasterInfo *header;
uint8         colormap[MAX_PAL], *pix;
{
  int i,j=0;
  
  uint8 *image = info->image;

  int   linesize;

  linesize = info->width;
  if (linesize % 2) linesize++;

  /* set up the header  */
  header->magic  = MAGIC;
  header->width  = info->width;
  header->height = info->height;
  header->depth  = 8;   /* 8-bits raster image */
  header->length = linesize * info->height;
  header->type   = RT_STANDARD;   /* No compression */
  header->maptype= RMT_EQUAL_RGB;    /* Color mode */
  header->maplen = MAX_PAL;

  j=0;
  if (info->values >> 1)  {     /* palette valid   */
    for (i=0; i<MAX_PAL/3; i++) {    /* rgb rgb rgb ...-> rrr...gggg....bbb  */
      colormap[i]               = info->palette[j++];
      colormap[MAX_PAL/3+i]     = info->palette[j++];
      colormap[(MAX_PAL*2)/3+i] = info->palette[j++];
    }
  } 
  else {
    for (i=0; i<MAX_PAL/3; i++) {
      colormap[i]                  = i;
      colormap[MAX_PAL/3+ i]       = i;
      colormap[(MAX_PAL * 2)/3 + i]= i;
    }
  }
 
  /* copy image to pix  */
  if (linesize == info->width) { 
    for (i=0; i<info->width*info->height;i++)
      *pix++ = *image++;
  }
  else {
    for (i=0; i<info->height;i++) {
       for (j=0; j<info->width; j++) 
	*pix++ = *image++;
       *pix++ = 0;
    }
  }

  return(AllOk);
}

/* write out the sun raster image */
ErrorCode writePixrect(info,header,colormap,image)
FileInfo      *info;
SunRasterInfo header;
uint8 colormap[MAX_PAL], *image;
{

  FILE  *fp;
  int   ret;
  int   linesize;

  linesize = info->width;
  if (linesize %2)  linesize++;

  /* Open the file */
    
  fp = fopen(info->name,"w");
  if (fp == NULL) return(err_msg(WriteSunRastFile, OutputOpenFailed));
    
  /* write info of the header to sunraster image */
  ret = fwrite(&header,sizeof(header),1,fp);
  if (ret != 1) return(err_msg(WriteSunRastFile, FileSystemError));

  /* write out the colormap */
  ret = fwrite(colormap, MAX_PAL, 1, fp);
  if (ret != 1) return(err_msg(WriteSunRastFile, FileSystemError));

  /* write out image */
  ret = fwrite(image, linesize*info->height, 1, fp);
  if (ret != 1) return(err_msg(WriteSunRastFile, FileSystemError));

  fclose(fp);

  return(Done);
}


ErrorCode WriteSunrast(info)
    FileInfo *info;
{   
    SunRasterInfo header;
    int32 width, height;
    int   depth,i,j ,row,col;
    uint8 *pix, *ptr;           /* The sun pixrect       */
    uint8 colormap[MAX_PAL];      /* The Pixrect Colormap  */
    uint8 red[256], green[256], blue[256];
    int   datasize;
    ErrorCode error;

    char *inp=NULL;
    char *outp=NULL;
    if (info == NULL) return(err_msg(WriteSunRastFile, NullPtr));
    
    if (info->width % 2) 
       datasize = (info->width+1) * info->height;
    else
       datasize = info->width*info->height;

    if ((pix = (uint8 *)HDmalloc(datasize)) == NULL)
      return(err_msg(WriteSunRastFile, NoMemory));

    ptr = pix;

    /* convert HDF RIS8 to the pixrect data and color map  */
    if ((error = HdfToPixrect(info, &header, colormap, pix))!= AllOk)
      return(error);

    /* write to sun raster file */
    error = writePixrect(info, header, colormap, ptr);

    /* free the space */
  
    free(ptr);

    return(error);

}
